/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file ncxparser.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 *
 * NCX parser for the Daisy Z39.86-2005 standard. Collects data and stores it in linked list
 * in the main datastructure. This linked smil list represents the spine of the
 * daisy DTB.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml2/libxml/xmlreader.h>

#include "ncxparser.h"
#include "control.h"
#include "report.h"
#include "common.h"
#include "snprintf/snprintf.h"

/* Test if the xml reader is present */
#ifdef LIBXML_READER_ENABLED



/* ************************************************** *
 * 			Static function declarations 			  *
 * ************************************************** */
 
 /**
 * Parses a daisy Z39.86-2005 ncx <head>...</head> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxHead (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <title>...</title> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxDocTitle (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <docauthor>...</docauthord> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxDocAuthor (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <navmap>...</navmap> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavMap (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <pagelist>...</pagelist> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxPageList (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <navlist>...</navlist> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavList (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <smilcustomtest>...</smilcutstomtest> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxSmilCustomTest (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <meta>...</meta> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxMeta (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <text>...</text> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @param mode - tells wheter we got a title (mode = 0), author (mode = 1) or a textpassage (mode = 2)
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxText (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode, int mode);

/**
 * Parses a daisy Z39.86-2005 ncx <audio>...</audio> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @param mode - tells wheter we got a title (mode = 0), author (mode = 1) or a textpassage (mode = 2)
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxAudio (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode, int mode);

/**
 * Parses a daisy Z39.86-2005 ncx <img>...</img> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param mode - tells wheter we got a title (mode = 0), author (mode = 1) or a textpassage (mode = 2)
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxImg (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode, int mode);

/**
 * Parses a daisy Z39.86-2005 ncx <navinfo>...</navinfo> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavInfo (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <navlabel>...</navlabel> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavLabel (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode);

/**
 * Parses a daisy Z39.86-2005 ncx <navpoint>...</navpoint> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavPoint (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <content>...</content> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxContent (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);

/**
 * Parses a daisy Z39.86-2005 ncx <pagetarget>...</pagetarget> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxPageTarget (xmlTextReaderPtr reader, struct DaisyData *daisydata);

/**
 * Parses a daisy Z39.86-2005 ncx <navtarget>...</navtarget> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavTarget (xmlTextReaderPtr reader, struct DaisyData *daisydata);



/* ************************************************** *
 * 			Global function definitions 			  *
 * ************************************************** */
 
/**
 * Function for parsing a Z39.86-2005 ncx file (*.ncx). Builds a linked list in the main
 * data structure containing information about all smil files. The linked list represents
 * the spine of the daisy DTB.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param filename - a pointer to a string containing the path and filename to the ncx file.
 * @returns 1 in case of success and -1 in case of error.
 */
int parseNCX (struct DaisyData *daisydata, char *filename)
{
 	const xmlChar *name;
 	char *nameChar = NULL;
 	xmlTextReaderPtr reader;
 	int ret;
 	
 	name = NULL;
 	/* open the ncx file */
 	reader = xmlReaderForFile (filename, NULL, 0);
 	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error),  "Failed to open NCX file: %s", filename);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCXPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX file: %s", filename);
	 		report (error, REP_DEBUG);
		}
 		
	 	/* read the first tag */
	 	ret = xmlTextReaderRead(reader);
	 	
	 	/* while there are more tags, parse the next */
	 	if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse NCX file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
	 	}
	 	
	 	/* Attributes:
	 	 * version
	 	 * xmlns
	 	 * xml:lang
	 	 * dir
	 	 */
	 	
	 	while(ret == 1)
	 	{
	 		/* get name of tag */
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return ret;
			nameChar = tolowercase((const char *)name);
			if (nameChar == NULL) return 0;
			/* parse tag if <head> */
			if (!strcmp (nameChar, "head")) ret = parseNcxHead (reader, daisydata);
			/* parse tag if <docTitle> */
			else if (!strcmp (nameChar, "doctitle")) ret = parseNcxDocTitle (reader, daisydata);
			/* parse tag if <docAuthor> * */
			else if (!strcmp (nameChar, "docauthor")) ret = parseNcxDocAuthor (reader, daisydata);
			/* parse tag if <navMap> */
			else if (!strcmp (nameChar, "navmap")) ret = parseNcxNavMap (reader, daisydata);
			/* parse tag if <pageList> ? */
			else if (!strcmp (nameChar, "pagelist")) ret = parseNcxPageList (reader, daisydata);
			/* parse tag if <navList> * */
			else if (!strcmp (nameChar, "navlist")) ret = parseNcxNavList (reader, daisydata);
			if (ret == -1) break;
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse NCX file: %s", filename);
 			report (error, REP_WARNING
 			);
 			return -1;
		}
	}
	return 1;
}


/* ************************************************** *
 * 			Static function definitions 			  *
 * ************************************************** */

 /**
 * Parses a daisy Z39.86-2005 ncx <head>...</head> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxHead (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;

	if (DEBUG_NCXPARSER) report ("Parsing NCX <HEAD>", REP_DEBUG);	
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <HEAD>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	/* while we haven't come to the head-close tag */
	while (strcmp (nameChar, "head"))
	{	
			/* parse tag if <smilCustomTest> */
			if (!strcmp (nameChar, "smilcustomtest")) ret = parseNcxSmilCustomTest (reader, daisydata);
			/* parse tag if <meta> */
			else if (!strcmp (nameChar, "meta")) ret = parseNcxMeta (reader, daisydata);
			if (ret != 1) return ret;
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);			
			if (name == NULL) return ret;
			nameChar = tolowercase((const char *)name);
			if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <HEAD>, endtag expected", REP_ERROR);
		return -1;
	}
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <smilcustomtest>...</smilcutstomtest> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxSmilCustomTest (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	int ret = 1;
	
	/* Attributes:
	 * id
	 * defaultState
	 * override
	 * bookStruct
	 */
	 
	 daisydata = daisydata;
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <SMILCUSTOMTEST>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <meta>...</meta> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxMeta (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	int ret = 1;
	xmlChar *attrvalue = NULL, *attrcontent = NULL;
	
	/* Attributes:
	 * name
	 * content
	 * scheme
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <META>", REP_DEBUG);
	attrvalue = xmlTextReaderGetAttribute (reader, (xmlChar *)"name");
	if (attrvalue != NULL)
	{
		attrcontent = xmlTextReaderGetAttribute (reader, (xmlChar *)"content");
		if (attrcontent != NULL)
		{	
			/* extract the title */
			if (!strcmp ((char *) attrvalue, "dc:Title") || !strcmp ((char *) attrvalue, "dtb:sourceTitle"))
			{	
				/* if we haven't parsed a title tag */
				if (daisydata->bookInfo->titleText == NULL)
				{
					daisydata->bookInfo->titleText = (char *) attrcontent;
				}
			}
			/* extract the total time of the book */
			else if (!strcmp ((char *) attrvalue, "dtb:totalTime"))
			{
				daisydata->bookInfo->totalTime = (char *) attrcontent;
			}						
		}
	}
	ret = xmlTextReaderRead (reader);
	return ret;	
}


/**
 * Parses a daisy Z39.86-2005 ncx <title>...</title> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxDocTitle (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;	
	
	/* Attributes:
	 * id
	 */
	 
	 daisydata = daisydata;
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <DOCTITLE>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <DOCTITLE>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	 
	/* while we haven't reached the end-tag of doctitle */
	while (strcmp (nameChar, "doctitle"))
	{	
		/* parse tag if <text> */
		if (!strcmp (nameChar, "text")) ret = parseNcxText (reader, daisydata, NULL, 0);
		/* parse tag if <audio> ? */
		else if (!strcmp (nameChar, "audio")) ret = parseNcxAudio (reader, daisydata, NULL, 0);
		/* parse tag if <img> ? */
		else if (!strcmp (nameChar, "img")) ret = parseNcxImg (reader, daisydata, NULL, 0);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <DOCTITLE>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <text>...</text> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @param mode - tells wheter we got a title (mode = 0), author (mode = 1) or a textpassage (mode = 2)
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxText (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode, int mode)
{
	xmlChar *value = NULL;
	char *tempChar = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <TEXT>", REP_DEBUG);
	
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	/* if the title tag has a value */
	if (xmlTextReaderHasValue (reader))
	{
		value = xmlTextReaderValue (reader);
		tempChar = removewhitespaces ((char *)value);
		/* storing the books title in the datastucture */
		if (mode == 0) { daisydata->bookInfo->titleText = tempChar; tempChar = NULL; }
		/* storing a authors name */
		else if (mode == 1) { daisydata->bookInfo->author->name = tempChar; tempChar = NULL; }
		/* storing a textpassage corresponding to the navLabel (navInfo not fully implemented) */
		else if (mode == 2)
		{
			if (tmpSmilNode != NULL)
			tmpSmilNode->textpassage = tempChar; tempChar = NULL;
		}
	}
	if (value != NULL) xmlFree (value); value = NULL;
	if (tempChar != NULL) free (tempChar); tempChar = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <audio>...</audio> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @param mode - tells wheter we got a title (mode = 0), author (mode = 1) or a textpassage (mode = 2)
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxAudio (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode, int mode)
{
	xmlChar *tmp = NULL;
	char *tempChar = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 * src
	 * clipBegin
	 * clipEnd
	 */

	if (DEBUG_NCXPARSER) report ("Parsing NCX <AUDIO>", REP_DEBUG); 
	/* fetch the url anchor to the audio file */
	tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen((char *)tmp)+1);
		strcpy(tempChar, (char *)tmp);
		if (DEBUG_NCXPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <AUDIO> Audiofile: %s", tmp);
			report (error, REP_DEBUG);
		}
		/* storing the audiofile corresponding to the title in the datastucture */
		if (mode == 0) { daisydata->bookInfo->titleAudiofilename = tempChar; tempChar = NULL; }
		/* storing a audiofile corresponding to the author */
		else if (mode == 1) { daisydata->bookInfo->author->audiofilename = tempChar; tempChar = NULL; }
		/* storing an audiofile textpassage corresponding to the navLabel (navInfo not fully implemented) */
		else if (mode == 2)
		{
			if (tmpSmilNode != NULL)
			tmpSmilNode->audiofilename = tempChar; tempChar = NULL;
		}		
		if (tempChar != NULL) free (tempChar); tempChar = NULL;
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
		
		/* fetch clipBegin */
		tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"clipBegin");
		if (tmp != NULL)
		{
			tempChar = getTime((char *)tmp);
			if (DEBUG_NCXPARSER)
			{
				char error[STRLEN];
				snprintf (error, sizeof (error), "Parsing NCX <AUDIO> clipBegin: %s", tempChar);
				report (error, REP_DEBUG);
			}
			/* storing the audiofile's startPosition corresponding to the title in the datastucture */
			if (mode == 0) { daisydata->bookInfo->titleAudioStartPos = tempChar; tempChar = NULL; }
			/* storing a audiofile's startPosition corresponding to the author */
			else if (mode == 1) { daisydata->bookInfo->author->audioStartPos = tempChar; tempChar = NULL; }
			/* storing the audiofiles's startPosition to the textpassage corresponding to the navLabel (navInfo not fully implemented) */
			else if (mode == 2)
			{
				if (tmpSmilNode != NULL)
				tmpSmilNode->audioStartPos = tempChar; tempChar = NULL;
			}
			if (tempChar != NULL) free (tempChar); tempChar = NULL;
			if (tmp != NULL) xmlFree (tmp); tmp = NULL;
						
			/* fetch clip-end */
			tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"clipEnd");
			if (tmp != NULL)
			{
				tempChar = getTime((char *)tmp);
				if (DEBUG_NCXPARSER)
				{
					char error[STRLEN];
					snprintf (error, sizeof (error), "Parsing NCXL <AUDIO> clipEnd: %s", tempChar);
					report (error, REP_DEBUG);
				}
				/* storing the audiofile's stopPosition corresponding to the title in the datastucture */
				if (mode == 0) { daisydata->bookInfo->titleAudioStopPos = tempChar; tempChar = NULL; }
				/* storing a audiofile's stopPosition corresponding to the author */
				else if (mode == 1) { daisydata->bookInfo->author->audioStopPos = tempChar; tempChar = NULL; }
				/* storing the audiofiles's stopPosition to the textpassage corresponding to the navLabel (navInfo not fully implemented) */
				else if (mode == 2)
				{
					if (tmpSmilNode != NULL)
					tmpSmilNode->audioStopPos = tempChar; tempChar = NULL;
				}
				if (tempChar != NULL) free (tempChar); tempChar = NULL;
				if (tmp != NULL) xmlFree (tmp); tmp = NULL;
			}
		}
		else
		{
			ret = 0;
		}
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp);	tmp = NULL;
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <img>...</img> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @param mode - tells wheter we got a title (mode = 0), author (mode = 1) or a textpassage (mode = 2)
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxImg (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode, int mode)
{
	xmlChar *tmp = NULL;
	char *tempChar;
	int ret = 1;

	/* Attributes:
	 * id
	 * class
	 * src
	 */

	if (DEBUG_NCXPARSER) report ("Parsing NCX <IMG>", REP_DEBUG);
	/* fetch the url anchor to the audio file */
	tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen((char *)tmp)+1);
		strcpy(tempChar, (char *)tmp);
		if (DEBUG_NCXPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <IMG> Imagefile: %s", tmp);
			report (error, REP_DEBUG);
		}		
		/* storing the image corresponding to the title in the datastucture */
		if (mode == 0) { daisydata->bookInfo->titleImage = tempChar; tempChar = NULL; }
		/* storing the image corresponding to the author */
		else if (mode == 1) { daisydata->bookInfo->author->image = tempChar; tempChar = NULL; }
		/* storing the image to the textpassage corresponding to the navLabel (navInfo not fully implemented) */
		else if (mode == 2)
		{
			if (tmpSmilNode != NULL)
			tmpSmilNode->image = tempChar; tempChar = NULL;
		}
		if (tempChar != NULL) free (tempChar); tempChar = NULL;
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp);	tmp = NULL;
	return ret;
}

/**
 * Parses a daisy Z39.86-2005 ncx <docauthor>...</docauthord> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxDocAuthor (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <DOCAUTHOR>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <DOCAUTHOR>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	(void) addNewAuthor (daisydata);
	/* while the current node is text or audio or img */
	while (strcmp (nameChar, "docauthor"))
	{	
		/* parse tag if <text> */
		if (!strcmp (nameChar, "text")) ret = parseNcxText (reader, daisydata, NULL, 1);
		/* parse tag if <audio> ? */
		else if (!strcmp (nameChar, "audio")) ret = parseNcxAudio (reader, daisydata, NULL, 1);
		/* parse tag if <img> ? */
		else if (!strcmp (nameChar, "img")) ret = parseNcxImg (reader, daisydata, NULL, 1);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <DOCAUTHOR>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <navmap>...</navmap> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavMap (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;
	
	name = NULL;
	/* Attributes:
	 * 
	 */
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVMAP>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVMAP>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	
	/* while we haven't reached the end-tag of navMap */
	while (strcmp (nameChar, "navmap"))
	{	
		/* parse tag if <navInfo> * */
		if (!strcmp (nameChar, "navinfo")) ret = parseNcxNavInfo (reader, daisydata);
		/* parse tag if <navLabel> * */
		else if (!strcmp (nameChar, "navlabel")) ret = parseNcxNavLabel (reader, daisydata, NULL);
		/* parse tag if <navPoint> + */
		else if (!strcmp (nameChar, "navpoint")) ret = parseNcxNavPoint (reader, daisydata);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVMAP>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}

/**
 * Parses a daisy Z39.86-2005 ncx <pagelist>...</pagelist> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxPageList (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 * class
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <PAGELIST>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <PAGELIST>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	
	/* while we haven't come to the end-tag of pageList */
	while (strcmp (nameChar, "pagelist"))
	{	
		/* parse tag if <navInfo> * */
		if (!strcmp (nameChar, "navinfo")) ret = parseNcxNavInfo (reader, daisydata);
		/* parse tag if <navLabel> * */
		else if (!strcmp (nameChar, "navlabel")) ret = parseNcxNavLabel (reader,daisydata, NULL);
		/* parse tag if <pageTarget> + */
		else if (!strcmp (nameChar, "pagetarget")) ret = parseNcxPageTarget (reader, daisydata);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <PAGELIST>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}

/**
 * Parses a daisy Z39.86-2005 ncx <navlist>...</navlist> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavList (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;

	/* Attributes:
	 * id
	 * class
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVLIST>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVLIST>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	
	/* while we haven't come to the end-tag of navList */
	while (strcmp (nameChar, "navlist"))
	{	
		/* parse tag if <navInfo> * */
		if (!strcmp (nameChar, "navinfo")) ret = parseNcxNavInfo (reader, daisydata);
		/* parse tag if <navLabel> + */
		else if (!strcmp (nameChar, "navlabel")) ret = parseNcxNavLabel (reader, daisydata, NULL);
		/* parse tag if <navTarget> + */
		else if (!strcmp (nameChar, "navtarget")) ret = parseNcxNavTarget (reader, daisydata);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}	
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVLIST>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <navinfo>...</navinfo> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavInfo (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{	
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * xml:lang
	 * dir
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVINFO>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVINFO>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	
	/* while we haven't reached the end-tag of navInfo */
	while (strcmp (nameChar, "navinfo"))
	{	
		/* parse tag if <text> */
		if (!strcmp (nameChar, "text")) ret = parseNcxText (reader, daisydata, NULL, 2);
		/* parse tag if <audio> (if <text> then <audio> ?) */
		else if (!strcmp (nameChar, "audio")) ret = parseNcxAudio (reader, daisydata, NULL, 2);
		/* parse tag if <img> ? */
		else if (!strcmp (nameChar, "img")) ret = parseNcxImg (reader, daisydata, NULL, 2);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVINFO>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <navlabel>...</navlabel> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavLabel (xmlTextReaderPtr reader, struct DaisyData *daisydata, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;

	/* Attributes:
	 * xml:lang
	 * dir
	 */
	
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVLABEL>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVLABEL>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	
	/* while we haven't reached the end-tag of navLabel */
	while (strcmp (nameChar, "navlabel"))
	{	
		/* parse tag if <text> */
		if (!strcmp (nameChar, "text")) ret = parseNcxText (reader, daisydata, tmpSmilNode, 2);
		/* parse tag if <audio> (if <text> then <audio> ?) */
		else if (!strcmp (nameChar, "audio")) ret = parseNcxAudio (reader, daisydata, tmpSmilNode, 2);
		/* parse tag if <img> ? */
		else if (!strcmp (nameChar, "img")) ret = parseNcxImg (reader, daisydata, tmpSmilNode, 2);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVLABEL>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <navpoint>...</navpoint> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavPoint (xmlTextReaderPtr reader,struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret = 1, type = 1;
	struct SmilNode *tmpSmilNode = NULL;
	
	/* Attributes:
	 * id
	 * class
	 * playOrder
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVPOINT>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVPOINT>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	
	/* make a new node in the linked list */
	tmpSmilNode = addNewSmilNode(daisydata);
	
	/* while we haven't reached the end-tag of navPoint	*/
	while (type != 15 && strcmp (nameChar, "navpoint"))
	{	
		/* parse tag if <navLabel> + */
		if (!strcmp (nameChar, "navlabel")) ret = parseNcxNavLabel (reader, daisydata, tmpSmilNode);
		/* parse tag if <content> */
		else if (!strcmp (nameChar, "content")) ret = parseNcxContent (reader, tmpSmilNode);
		/* parse tag if <navPoint> * */
		else if (!strcmp (nameChar, "navpoint")) ret = parseNcxNavPoint (reader, daisydata);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
		type = xmlTextReaderNodeType(reader);
	}	
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <content>...</content> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param tmpSmilNode - a pointer to a new Smilnode in the linked list for storing the passage data.
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxContent (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	xmlChar *smilsrc = NULL;
	char *tempChar = NULL, *tmp = NULL;
	int ret = 1;

	/* Attributes:
	 * id
	 * src
	 */

	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVCONTENT>", REP_DEBUG);
	/* gets the href attribute. */
	smilsrc = xmlTextReaderGetAttribute (reader, (xmlChar *)"src");
	if (smilsrc != NULL)
	{
		/* splits the text on #. */
		tmp = strtok ((char *)smilsrc, "#");
		tempChar = (char *) malloc (strlen (tmp)+1);
		strcpy(tempChar, tmp);
		
		if (DEBUG_NCXPARSER)
		{ 
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <CONTENT> File anchor: %s", tmp);
			report (error, REP_DEBUG);
		}
		
		/* put the anchor in the node */
		if (tmpSmilNode != NULL)
		tmpSmilNode->anchor = tempChar; tempChar = NULL;
		
		/* get the fragment identifier */
		tmp = strtok (NULL, "#");
		tempChar = (char *) malloc (strlen (tmp)+1);
		strcpy(tempChar, tmp);
		
		if (DEBUG_NCXPARSER)
		{ 
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <CONTENT> Fragment Identifier: %s", tmp);
			report (error, REP_DEBUG);
		}
		
		/* put the fragment identifier in the node */
		if (tmpSmilNode != NULL)
		tmpSmilNode->fragmentIdentifier = tempChar; tempChar = NULL;
		
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		if (smilsrc != NULL) xmlFree(smilsrc); smilsrc = NULL;
	}
	else
	{
		ret = 0;
	}
	return ret;
}

/**
 * Parses a daisy Z39.86-2005 ncx <pagetarget>...</pagetarget> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxPageTarget (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 * value
	 * type
	 * class
	 * playOrder
	 */
	
	if (DEBUG_NCXPARSER) report ("Parsing NCX <PAGETARGET>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <PAGETARGET>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;
	
	/* while we haven't come to the end tag of pageTarget */
	while (!strcmp (nameChar, "pagetarget"))
	{	
		/* parse tag if <navLabel> + */
		if (!strcmp (nameChar, "navlabel")) ret = parseNcxNavLabel (reader, daisydata, NULL);
		/* parse tag if <content> */
		else if (!strcmp (nameChar, "content")) ret = parseNcxContent (reader, NULL);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <PAGETARGET>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parses a daisy Z39.86-2005 ncx <navtarget>...</navtarget> tag.
 * @param reader - the XmlTextReader opened to parse the ncx file.
 * @param daisydata - a pointer to the datastructure initialized by parseInit ().
 * @returns 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavTarget (xmlTextReaderPtr reader, struct DaisyData *daisydata)
{
	const xmlChar *name = NULL;
	char *nameChar = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 * value
	 * class
	 * playOrder
	 */
	
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVTARGET>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVTARGET>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	nameChar = tolowercase((const char *)name);
	if (nameChar == NULL) return 0;

	/* while we haven't come to the end tag of navTarget */
	while (strcmp (nameChar, "navtarget"))
	{
		/* parse tag if <navLabel> + */
		if (!strcmp (nameChar, "navlabel")) ret = parseNcxNavLabel (reader, daisydata, NULL);
		/* parse tag if <content> */
		else if (!strcmp (nameChar, "content")) ret = parseNcxContent (reader, NULL);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		nameChar = tolowercase((const char *)name);
		if (nameChar == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCX <NAVTARGET>, endtag expected", REP_ERROR);
		return -1;
	}		
	return ret;
}


#endif
